from qiskit.circuit import QuantumCircuit, CircuitInstruction
from qiskit.quantum_info import Statevector
from qiskit.circuit import InstructionSet
from typing import Generator, Iterable, Tuple, List, Set, Optional
import numpy as np


def group_instructions(qc: QuantumCircuit) -> List[List[CircuitInstruction]]:
    """
    Group instructions in a quantum circuit based on qubit dependencies.

    This function groups quantum circuit instructions such that instructions operating
    on the same qubits are placed in the same group. Instructions are separated into
    new groups when:
    1. A barrier instruction is encountered
    2. An instruction operates on qubits that are already active in the current group

    Parameters:
    -----------
    qc: QuantumCircuit
        The quantum circuit to group instructions from

    Returns:
    --------
    List[List[CircuitInstruction]]
        A list of instruction groups, where each group contains instructions that can be
        executed in parallel (i.e., they operate on different qubits)

    Raises:
    -------
    ValueError
        If the quantum circuit is empty or invalid
    """
    if not qc or not qc.data:
        return []

    current_group: List[CircuitInstruction] = []
    active_qubits: Set[str] = set()
    instruction_groups: List[List[CircuitInstruction]] = []

    def get_qubit_identifier(qubit) -> str:
        """Helper function to get a unique string identifier for a qubit."""
        return f"{qubit._register.prefix}{qubit._index}"

    for instruction in qc.data:
        operation_name = instruction.operation.name

        # Handle barrier instructions
        if operation_name == 'barrier':
            if current_group:
                instruction_groups.append(current_group)
                current_group = []
                active_qubits.clear()
            continue

        # Skip measurement instructions
        if operation_name == 'measure':
            continue

        # Get qubits involved in the current instruction
        qubits_involved = {get_qubit_identifier(q) for q in instruction.qubits}

        # Check if any qubit is already active in the current group
        if active_qubits & qubits_involved:
            instruction_groups.append(current_group)
            current_group = []
            active_qubits.clear()

        # Add instruction to current group and update active qubits
        current_group.append(instruction)
        active_qubits.update(qubits_involved)

    # Add the last group if it's not empty
    if current_group:
        instruction_groups.append(current_group)

    return instruction_groups


def generate_states(
    qc: QuantumCircuit,
    initial_state: Statevector,
    grouped_instructions: Optional[List[List[CircuitInstruction]]] = None
) -> Generator[Tuple[Statevector, Iterable[InstructionSet]], None, None]:
    """
    Generate the states of the quantum circuit at each step.

    This function simulates the evolution of the quantum state through the circuit,
    yielding the state and the operations that led to it at each step.

    Parameters:
    -----------
    qc: QuantumCircuit
        The quantum circuit to analyze
    initial_state: Statevector
        The initial state of the quantum circuit
    grouped_instructions: Optional[List[List[CircuitInstruction]]]
        Pre-grouped instructions to use. If None, instructions will be grouped.

    Yields:
    -------
    Tuple[Statevector, Iterable[InstructionSet]]
        A tuple containing:
        - The current quantum state
        - The operations that led to this state

    Raises:
    -------
    ValueError
        If the initial state dimension doesn't match the circuit
    """
    if initial_state.dim != 2**qc.num_qubits:
        raise ValueError(
            f"Initial state dimension {initial_state.dim} doesn't match circuit size {2**qc.num_qubits}"
        )

    state = initial_state
    if grouped_instructions is None:
        grouped_instructions = group_instructions(qc)

    for instructionSet in grouped_instructions:
        yield (state, instructionSet)

        # Create a sub-circuit for the current instruction group
        sub_circuit = QuantumCircuit(*qc.qregs, *qc.cregs)
        for instruction in instructionSet:
            sub_circuit.append(
                instruction.operation,
                instruction.qubits,
                instruction.clbits
            )

        state = state.evolve(sub_circuit)

    # Yield the final state
    yield (state, [])


def interpolate_states(
    qc: QuantumCircuit,
    initial_state: Statevector,
    intermediate_steps: int,
    grouped_instructions: Optional[List[List[CircuitInstruction]]] = None
) -> Generator[Tuple[Statevector, Iterable[InstructionSet]], None, None]:
    """
    Interpolate between two adjacent states to smooth the transition.

    This function creates intermediate states between each pair of adjacent states
    in the circuit evolution, allowing for smoother visualization of the state
    changes.

    Parameters:
    -----------
    qc: QuantumCircuit
        The quantum circuit to analyze
    initial_state: Statevector
        The initial state of the quantum circuit
    intermediate_steps: int
        The number of intermediate steps to interpolate between two adjacent states
    grouped_instructions: Optional[List[List[CircuitInstruction]]]
        Pre-grouped instructions to use. If None, instructions will be grouped.

    Yields:
    -------
    Tuple[Statevector, Iterable[InstructionSet]]
        A tuple containing:
        - The interpolated quantum state
        - The operations that led to this state

    Raises:
    -------
    ValueError
        If intermediate_steps is less than 1
    """
    if intermediate_steps < 1:
        raise ValueError("intermediate_steps must be at least 1")

    if grouped_instructions is None:
        grouped_instructions = group_instructions(qc)

    states = list(generate_states(qc, initial_state, grouped_instructions))

    # Pre-compute weights and their complements for better performance
    weights = np.linspace(0, 1, intermediate_steps, endpoint=False)
    weights_complement = 1 - weights

    # Interpolate between each pair of adjacent states
    for lhs, rhs in zip(states[:-1], states[1:]):
        lhs_data = lhs[0].data
        rhs_data = rhs[0].data

        # Vectorized interpolation for all intermediate steps at once
        intermediate_states = np.outer(weights_complement, lhs_data) + np.outer(weights, rhs_data)

        # Normalize all states at once
        norms = np.linalg.norm(intermediate_states, axis=1, keepdims=True)
        intermediate_states = np.divide(intermediate_states, norms, where=(norms > 0))

        # Yield each normalized intermediate state
        for state in intermediate_states:
            yield (Statevector(state), lhs[1])

    # Yield the final state
    yield states[-1]
